#![allow(clippy::manual_strip)]

use std::array;
use std::env;
use std::fs::{self, File};
use std::io::{BufWriter, Write};
use std::process::ExitCode;

const PACK_SIZE: usize = 24;
const PACKS_PER_SECTOR: usize = 4;
const SECTOR_SIZE: usize = PACK_SIZE * PACKS_PER_SECTOR;
const SECTOR_SPREAD: usize = 2;
const CHANNEL_SIZE: usize = SECTOR_SIZE / 8;

type Sector = [u8; SECTOR_SIZE];

use gf256::{gf, rs::rs};

#[gf(polynomial = 0x43, generator = 0x2)]
type gf64;

#[rs(gf=gf64, u=u8, block=24, data=20)]
mod p_parity {}

#[rs(gf=gf64, u=u8, block=4, data=2)]
mod q_parity {}

fn main() -> ExitCode {
    let args: Vec<_> = env::args_os().collect();
    if args.len() != 3 {
        eprintln!("usage: cdg-to-sub in.cdg out.sub");
        return ExitCode::FAILURE;
    }
    let mut in_file = fs::read(&args[1]).expect("read input");
    if in_file.len() % SECTOR_SIZE != 0 {
        eprintln!("ERROR: input must be a multiple of {SECTOR_SIZE} bytes");
        return ExitCode::FAILURE;
    }
    let in_file_sector_count = in_file.len() / SECTOR_SIZE;

    let mut out_file = BufWriter::new(File::create(&args[2]).expect("open output"));

    for (pack_i, pack) in in_file.chunks_mut(PACK_SIZE).enumerate() {
        // ensure we don't have P-Q data, which is unsupported
        if let Some(i) = pack.iter().position(|b| *b & 0xc0 != 0) {
            eprintln!(
                "ERROR: Input must have no P-Q channel data, non-zero found at {:#x}",
                pack_i * PACK_SIZE + i
            );
            return ExitCode::FAILURE;
        }

        // warn if there is early data, which will be lost
        let sector_i = pack_i / PACKS_PER_SECTOR;
        if sector_i < SECTOR_SPREAD && !pack.iter().all(|b| *b == 0) {
            eprintln!(
                "WARNING: Non-zero bytes in early sector {sector_i} will be lost due to delay"
            );
        }

        // write parity
        q_parity::encode(&mut pack[0..4]);
        p_parity::encode(pack);
    }

    // pad to flush delay
    in_file.extend([0; SECTOR_SIZE * SECTOR_SPREAD]);

    for sector_i in 0..in_file_sector_count {
        // collect output
        let out_sector: [u8; SECTOR_SIZE] = array::from_fn(|i| {
            let pack_i = i / PACK_SIZE;
            let scrambled_col = i % PACK_SIZE;

            // scramble
            let unscrambled_col = match scrambled_col {
                1 => 18,
                18 => 1,

                2 => 5,
                5 => 2,

                3 => 23,
                23 => 3,

                _ => scrambled_col,
            };

            // convolutional interleaver delay
            let pack_delay = scrambled_col & 7;

            // pull forward delayed bytes
            let pack_start_offset = ((sector_i + SECTOR_SPREAD) * PACKS_PER_SECTOR + pack_i)
                .checked_sub(pack_delay)
                .expect("delay should be less than sector spread")
                * PACK_SIZE;
            in_file[pack_start_offset + unscrambled_col]
        });

        // pack channels
        let packed_sector: Sector = array::from_fn(|i| {
            let channel = i / CHANNEL_SIZE;
            let byte_i = i % CHANNEL_SIZE;

            let mut packed_byte = 0u8;
            for bit_i in 0..8 {
                let bit = (out_sector[byte_i * 8 + bit_i] << channel) & 0x80;
                packed_byte |= bit >> bit_i;
            }
            packed_byte
        });
        out_file
            .write_all(&packed_sector)
            .expect("write output sector");
    }

    out_file.flush().expect("flush output");

    ExitCode::SUCCESS
}
